{%- set widget = view %}
{%- set view = widget.options.view %}

{%- for action in view.object_actions %}
  {{ action.render() }}
{%- endfor %}

{%- for panel, data in panels %}
<div class="panel" style="margin-top: 10px;">
  <h2>{{ panel.label }}</h2>

  <table class="table table-striped table-condensed" style="margin: 10px 0 15px 0;">
    <tbody>
    {%- for row in panel %}
    {%- if row.cols[0] in data %}
      {%- set label, value = data[row.cols[0]] %}
      <tr class="row">
        <td class="col-sm-4">{{ label }}</td>
        <td class="col-sm-8" colspan="3">{{ value }}</td>
      </tr>
    {%- endif %}
    {%- endfor %}
    </tbody>
  </table>
</div>
{% endfor %}
